import os, os.path
from glob import glob
import platform
import sys
import shutil
import subprocess
import re
import sysconfig

from distutils import log
from setuptools.command.install import install
from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext
from distutils.version import LooseVersion

if os.environ.get('KIWI_CPU_ARCH'):
    from sysconfig import get_platform
    fd = get_platform().split('-')
    if fd[0] == 'macosx':
        os.environ['_PYTHON_HOST_PLATFORM'] = '-'.join(fd[:-1] + [os.environ['KIWI_CPU_ARCH']])

is_gil_free = sysconfig.get_config_var('Py_GIL_DISABLED')
is_limited_api = not is_gil_free and os.environ.get('Py_LIMITED_API') == '1'

def get_extra_cmake_options():
    """read --clean, --no, --set, --compiler-flags, and -G options from the command line and add them as cmake switches.
    """
    _cmake_extra_options = ["-DKIWI_BUILD_TEST=0", "-DCMAKE_POSITION_INDEPENDENT_CODE=1", "-DKIWI_USE_MIMALLOC=" + ("1" if os.environ.get('USE_MIMALLOC') else "0")]
    if os.environ.get('KIWI_CPU_ARCH'):
        _cmake_extra_options.append("-DKIWI_CPU_ARCH=" + os.environ['KIWI_CPU_ARCH'])
    if os.environ.get('MACOSX_DEPLOYMENT_TARGET'):
        _cmake_extra_options.append("-DCMAKE_OSX_DEPLOYMENT_TARGET=" + os.environ['MACOSX_DEPLOYMENT_TARGET'])
    if is_gil_free:
        _cmake_extra_options.append("-DPy_GIL_DISABLED=1")
    if is_limited_api:
        _cmake_extra_options.append("-DPy_LIMITED_API=0x03090000")
    _clean_build_folder = False
    print(_cmake_extra_options)

    opt_key = None

    argv = [arg for arg in sys.argv]  # take a copy
    # parse command line options and consume those we care about
    for arg in argv:
        if opt_key == 'compiler-flags':
            _cmake_extra_options.append('-DCMAKE_CXX_FLAGS={arg}'.format(arg=arg.strip()))
        elif opt_key == 'G':
            _cmake_extra_options += ['-G', arg.strip()]
        elif opt_key == 'no':
            _cmake_extra_options.append('-D{arg}=no'.format(arg=arg.strip()))
        elif opt_key == 'set':
            _cmake_extra_options.append('-D{arg}'.format(arg=arg.strip()))

        if opt_key:
            sys.argv.remove(arg)
            opt_key = None
            continue

        if arg == '--clean':
            _clean_build_folder = True
            sys.argv.remove(arg)
            continue

        if arg == '--yes':
            print("The --yes options to kiwipiepy's setup.py don't do anything since all these options ")
            print("are on by default.  So --yes has been removed.  Do not give it to setup.py.")
            sys.exit(1)
        if arg in ['--no', '--set', '--compiler-flags']:
            opt_key = arg[2:].lower()
            sys.argv.remove(arg)
            continue
        if arg in ['-G']:
            opt_key = arg[1:]
            sys.argv.remove(arg)
            continue

    return _cmake_extra_options, _clean_build_folder

def num_available_cpu_cores(ram_per_build_process_in_gb):
    import multiprocessing
    try:
        mem_bytes = os.sysconf('SC_PAGE_SIZE') * os.sysconf('SC_PHYS_PAGES')  
        mem_gib = mem_bytes/(1024.**3)
        num_cores = multiprocessing.cpu_count() 
        mem_cores = int(mem_gib/float(ram_per_build_process_in_gb)+0.5)
        return max(min(num_cores, mem_cores), 1)
    except ValueError:
        return 2 # just assume 2 if we can't get the os to tell us the right answer.

cmake_extra_options, clean_build_folder = get_extra_cmake_options()


class CMakeExtension(Extension):
    def __init__(self, name, sourcedir='', *args, **kwargs):
        Extension.__init__(self, name, sources=[], *args, **kwargs)
        self.sourcedir = os.path.abspath(sourcedir)

class CMakeBuild(build_ext):

    def get_cmake_version(self):
        try:
            out = subprocess.check_output(['cmake', '--version'])
        except:
            sys.stderr.write("\nERROR: CMake must be installed to build kiwipiepy\n\n") 
            sys.exit(1)
        return re.search(r'version\s*([\d.]+)', out.decode()).group(1)

    def run(self):
        cmake_version = self.get_cmake_version()
        if platform.system() == "Windows":
            if LooseVersion(cmake_version) < '3.1.0':
                sys.stderr.write("\nERROR: CMake >= 3.1.0 is required on Windows\n\n")
                sys.exit(1)

        for ext in self.extensions:
            self.build_extension(ext)

    def build_extension(self, ext):
        extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))
        libs = self.get_libraries(ext)
        if is_gil_free:
            for i, lib in enumerate(libs):
                if re.fullmatch(r'python3[0-9]+', lib):
                    libs[i] = lib + 't'
        if is_limited_api:
            for i, lib in enumerate(libs):
                if re.fullmatch(r'python3[0-9]+', lib):
                    libs[i] = 'python3'

        cmake_args = [
            '-DINCLUDE_DIRS={}'.format(';'.join(self.include_dirs)),
            '-DLIBRARY_DIRS={}'.format(';'.join(self.library_dirs)),
            '-DLIBRARIES={}'.format(';'.join(libs)),
            '-DPYTHON_EXECUTABLE=' + sys.executable,
        ]
        cmake_args += cmake_extra_options
        print(cmake_args)

        cfg = 'Debug' if self.debug else 'Release'
        build_args = ['--config', cfg]

        if platform.system() == "Windows":
            if sys.maxsize > 2**32:
                cmake_args += ['-A', 'x64']
            else:
                cmake_args += ['-A', 'Win32']
            build_args += ['--', '/m'] 
        else:
            cmake_args += ['-DCMAKE_BUILD_TYPE=' + cfg]
            build_args += ['--', '-j'+str(num_available_cpu_cores(2))]

        build_folder = os.path.abspath(self.build_temp)

        if not os.path.exists(build_folder):
            os.makedirs(build_folder)

        cmake_setup = ['cmake', ext.sourcedir] + cmake_args
        cmake_build = ['cmake', '--build', '.'] + build_args

        print("Building extension for Python {}".format(sys.version.split('\n',1)[0]))
        print("Invoking CMake setup: '{}'".format(' '.join(cmake_setup)))
        sys.stdout.flush()
        subprocess.check_call(cmake_setup, cwd=build_folder)
        print("Invoking CMake build: '{}'".format(' '.join(cmake_build)))
        sys.stdout.flush()
        subprocess.check_call(cmake_build, cwd=build_folder)
        
        fullname = self.get_ext_fullname(ext.name)
        filename = self.get_ext_filename(fullname)

        if platform.system() == "Windows":
            shutil.move(os.sep.join([build_folder, cfg, '_kiwipiepy.dll']), os.sep.join([extdir, filename]))
        else:
            shutil.move(glob(os.sep.join([build_folder, 'lib_kiwipiepy.*']))[0], os.sep.join([extdir, filename]))
        #print()


here = os.path.abspath(os.path.dirname(__file__))
exec(open('kiwipiepy/_version.py').read())

long_description = '''kiwipiepy
----------
kiwipiepy is a python version package of Kiwi(Korean Intelligent Word Identifier) which is a morphological analyzer for Korean.

https://github.com/bab2min/kiwipiepy '''


libraries = []

if platform.system() == 'Windows': 
    if os.environ.get('USE_MIMALLOC'): libraries.append('advapi32.lib')
else: 
    pass

setup(
    name='kiwipiepy',
    version=__version__,

    description='Kiwi, the Korean Tokenizer for Python',
    long_description=long_description,

    url='https://github.com/bab2min/kiwipiepy',

    author='bab2min',
    author_email='bab2min@gmail.com',

    license='LGPL v3 License',

    classifiers=[
        'Development Status :: 3 - Alpha',

        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Topic :: Software Development :: Libraries",
        "Topic :: Text Processing :: Linguistic",

        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",

        'Programming Language :: Python :: 3',
        'Programming Language :: C++'
    ],

    keywords='Korean morphological analysis',
    install_requires=[
        'dataclasses; python_version < "3.7"',
        'kiwipiepy_model>=0.22,<0.23',
        'numpy<2; python_version < "3.9"',
        'numpy; python_version >= "3.9"',
        'tqdm',
    ],
    packages=['kiwipiepy'],
    include_package_data=True,
    ext_modules=[CMakeExtension('_kiwipiepy',
        libraries=libraries,
        py_limited_api=is_limited_api,
    )],
    options={"bdist_wheel": {"py_limited_api": "cp39"}} if is_limited_api else {},
    cmdclass=dict(build_ext=CMakeBuild),
)



'''
.. versionadded:: 0.15.1

`sw_tokenizer` 모듈은 서브워드 토크나이저와 관련된 클래스를 제공합니다.
'''

import re
import itertools
from typing import Callable, List, Optional, Tuple, Union, Iterable, Dict, Any
from dataclasses import dataclass
import warnings

import tqdm

from _kiwipiepy import _SwTokenizer

from kiwipiepy import Kiwi, Token

@dataclass
class SwTokenizerConfig:
    '''
    서브워드 토크나이저의 설정을 관리하는 데이터 클래스입니다.
    '''
    
    lowercase:bool = False
    '''토큰화에 앞서 대문자들을 전부 소문자로 정규화합니다.'''
    split_chinese:bool = True
    '''토큰화에 앞서 모든 한자를 한 글자씩 분리합니다.'''
    whole_word_unk:bool = False
    '''토크나이저의 어휘 집합에 속하지 않은 글자가 하나라도 포함된 어절 전체를 UNK토큰으로 지정합니다.'''
    integrate_allomorph: bool = True
    '''이형태를 통합합니다. (아직 미지원)'''
    split_punct: bool = True
    '''토큰화에 앞서 구두점을 분리합니다.'''
    simple_tag:bool = True
    '''세부 품사태그(예: VV, VX, VA) 대신 대표 품사태그(예: V)를 사용합니다. '''
    split_verb:bool = True
    '''서브워드로 분리가능한 동사가 있다면 더 작은 단위로 분리합니다. (예: 뛰어가다 -> 뛰/V 어/E 가/V 다/E)'''
    split_eomi:bool = True
    '''서브워드로 분리가능한 어미가 있다면 더 작은 단위로 분리합니다. (예: 어요 -> 어/E 요/J)'''
    use_glue_token:bool = True
    '''단어를 이어붙이는 특수 토큰인 Glue토큰을 사용합니다.'''
    use_newline_token:bool = False
    '''줄바꿈 문자를 토큰으로 사용합니다. False일 경우 공백으로 취급하여 무시합니다.
    줄바꿈 문자로는 fallback_byte 토큰의 10번째 바이트 토큰(\n)을 사용합니다.
    '''
    strict:bool = False
    '''토큰화 수행시 원문을 완벽하게 복원하도록 합니다. (아직 미지원)'''
    fallback_hangul:bool = True
    '''토크나이저의 어휘 집합에 속하지 않은 한글이 있다면 초성+중성 / 종성의 형태로 분해하여 처리합니다.'''
    fallback_byte:bool = False
    '''토크나이저의 어휘 집합에 속하지 않은 문자가 있다면 UTF8 Byte로 변환하여 처리합니다.'''

    unk_token:str = "[UNK]"
    cls_token:str = None
    sep_token:str = None
    pad_token:str = None
    mask_token:str = None
    bos_token:str = None
    eos_token:str = None

    additional: Any = None

class TrainerCallback:
    '''
    서브워드 토크나이저를 학습할 때 진행상황을 보고받기 위한 콜백 클래스입니다.
    이 클래스를 상속 받아서 필요한 기능을 구현한 뒤 `SwTokenizer.train` 함수의 `callback` 인자로 넘겨 사용할 수 있습니다.
    '''
    def begin_tokenization(self, num_processed_lines:int):
        '''
이 메소드는 학습을 위해 형태소 분석 작업이 시작됐을 때 호출됩니다.

Parameters
----------
num_processed_lines: int
    처리가 완료된 줄 수
        '''

    def proc_tokenization(self, num_processed_lines:int):
        '''
이 메소드는 형태소 분석 작업이 진행 중일 때 호출됩니다.
인자는 `begin_tokenization`과 동일합니다.
        '''

    def end_tokenization(self, num_processed_lines:int):
        '''
이 메소드는 형태소 분석 작업이 완료되었을 때 호출됩니다.
인자는 `begin_tokenization`과 동일합니다.
        '''

    def begin_reduction(self, n_tkn:int, iteration:int, cur_vocab_size:int, unigram_loss:float):
        '''
이 메소드는 형태소 분석 작업 완료 후 토큰 후보를 추리는 작업이 시작됐을 때 호출됩니다.

Parameters
----------
n_tkn: int
    현재 학습 중인 토크나이저의 순번

iteration: int
    현재 완료된 반복 횟수

cur_vocab_size: int
    현재 토큰 집합의 크기

unigram_loss: float
    손실함수의 현재 값
        '''

    def proc_reduction(self, n_tkn:int, iteration:int, cur_vocab_size:int, unigram_loss:float):
        '''
이 메소드는 토큰 후보를 추리는 작업이 진행 중일 때 호출됩니다.
인자는 `begin_reduction`과 동일합니다.
        '''

    def end_reduction(self, n_tkn:int, iteration:int, cur_vocab_size:int, unigram_loss:float):
        '''
이 메소드는 토큰 후보를 추리는 작업이 완료되었을 때 호출됩니다.
인자는 `begin_reduction`과 동일합니다.
        '''

class _ProgressShower(TrainerCallback):
    def __init__(self, file, total=None, iterations=None):
        super().__init__()
        self._file = file
        self._bar = None
        self._total = total
        self._iterations = iterations
    
    def __del__(self):
        if self._bar:
            self._bar.close()

    def begin_tokenization(self, num_processed_lines: int):
        self._bar = tqdm.tqdm(itertools.repeat(None), desc="Tokenizing", file=self._file, total=self._total)
        self._last_proc_lines = 0
    
    def proc_tokenization(self, num_processed_lines: int):
        self._bar.update(num_processed_lines - self._last_proc_lines)
        self._last_proc_lines = num_processed_lines
    
    def end_tokenization(self, num_processed_lines: int):
        self._bar.update(num_processed_lines - self._last_proc_lines)
        self._last_proc_lines = num_processed_lines
        self._bar.close()
        self._bar = None
    
    def begin_reduction(self, n_tkn:int, iteration: int, cur_vocab_size: int, unigram_loss: float):
        self._bar = tqdm.tqdm(itertools.repeat(None), desc=f"Reducing #{n_tkn+1}", file=self._file, total=self._iterations)
        self._bar.write(f"Iteration: {iteration} VocabSize: {cur_vocab_size} Loss: {-unigram_loss:.4f}")
        self._bar.set_postfix(dict(vocab_size=cur_vocab_size, loss=-unigram_loss))
        self._last_iteration = iteration
    
    def proc_reduction(self, n_tkn:int, iteration: int, cur_vocab_size: int, unigram_loss: float):
        self._bar.write(f"Iteration: {iteration} VocabSize: {cur_vocab_size} Loss: {-unigram_loss:.4f}")
        self._bar.update(iteration - self._last_iteration)
        self._bar.set_postfix(dict(vocab_size=cur_vocab_size, loss=-unigram_loss))
        self._last_iteration = iteration

    def end_reduction(self, n_tkn:int, iteration: int, cur_vocab_size: int, unigram_loss: float):
        self._bar.close()
        self._bar = None
        print(f"Finished. Iteration: {iteration} VocabSize: {cur_vocab_size} Loss: {-unigram_loss:.4f}", file=self._file)

SPECIAL_TOKEN_NAMES = ['unk', 'cls', 'sep', 'mask', 'pad', 'bos', 'eos']

class SwTokenizer(_SwTokenizer):
    '''
형태소 분석기 Kiwi를 기반으로 한 서브워드 토크나이저(Subword Tokenizer)를 제공하는 클래스입니다.
일반 서브워드 토크나이저와는 달리 한국어의 형태소를 고려하며 분할을 수행한다는 특징이 있습니다.

Parameters
----------
path: str
    불러들일 토크나이저 파일의 경로입니다.

kiwi: Kiwi
    형태소 분석에 사용될 Kiwi 형태소 분석기의 인스턴스입니다.
    생략하거나 None으로 줄 경우 내부에서 자체적으로 Kiwi 인스턴스를 생성하여 사용합니다.

num_workers: int
    멀티스레딩 시 사용할 스레드의 개수입니다.
    `kiwi`를 생략한 경우에만 유효합니다.
    `kiwi`를 지정한 경우 `kiwi.num_workers` 값에 의해 멀티스레딩을 수행합니다.
    '''

    def __init__(self,
        path: str,
        kiwi: Optional[Kiwi] = None,
        num_workers: Optional[int] = None,
    ):
        if kiwi is None:
            kiwi = Kiwi(num_workers=num_workers, load_multi_dict=False)
        elif num_workers is not None:
            raise ValueError("You cannot specify `num_workers` value if you give `kiwi` value.")

        if not isinstance(kiwi, Kiwi):
            raise ValueError("`kiwi` must be an instance of `Kiwi`.")

        super().__init__(kiwi, path)

        self._space_tolerance = (self.config.additional.get('space_tolerance') if isinstance(self.config.additional, dict) else None) or 0
    
    def encode(self, 
        text: Union[str, Iterable[str]],
        return_offsets: bool = False,
    ) -> Union[List[int], Tuple[List[int], List[Tuple[int, int]]]]:
        '''
주어진 텍스트를 토큰화하여 token id의 리스트로 반환합니다.

Parameters
----------
text: Union[str, Iterable[str]]
    분할할 텍스트, 혹은 텍스트의 iterable
    이 인자를 단일 str로 줄 경우, 싱글스레드에서 처리하며
    str의 Iterable로 줄 경우, 멀티스레드로 분배하여 처리합니다.

return_offsets: bool
    True일 경우 각 토큰들의 텍스트 상의 시작지점 및 끝지점이 함께 반환됩니다.

Returns
-------
token_ids: List[int]
    `text`를 단일 `str`으로 주고 `return_offsets = False`인 경우.
    token id의 목록을 list로 반환합니다.

iterable_of_token_ids: Iterable[List[int]]
    `text`를 `Iterable[str]`으로 주고 `return_offsets = False`인 경우.
    token id의 list를 나열하는 Iterable을 반환합니다.

token_ids_and_offsets: Tuple[List[int], List[Tuple[int, int]]]
    `text`를 단일 `str`으로 주고 `return_offsets = True`인 경우.
    token id의 리스트와 각 토큰들의 시작지점과 끝지점(문자 단위)을 나타내는 tuple의 리스트를 반환합니다.

iterable_of_token_ids_and_offsets: Iterable[Tuple[List[int], List[Tuple[int, int]]]]
    `text`를 `Iterable[str]`으로 주고 `return_offsets = True`인 경우.

Notes
-----
멀티스레딩은 SwTokenizer를 생성할 때 준 kiwi에 의해 수행됩니다.
따라서 kiwi 인자로 준 Kiwi 인스턴스의 num_workers 값에 따라 동시에 수행되는 작업의 개수가 결정됩니다.

```python
from kiwipiepy import Kiwi
from kiwipiepy.sw_tokenizer import SwTokenizer
kiwi = Kiwi(num_workers=1)
tokenizer = SwTokenizer('some_tokenizer.json', kiwi=kiwi)
tokenizer.encode(...) 
# kiwi.num_workers가 1이기 때문에 1개 스레드에서만 작업이 처리됨
tokenizer = SwTokenizer('some_tokenizer.json', num_workers=1)
tokenizer.encode(...) 
# 위와 동일하게 1개 스레드 사용

kiwi = Kiwi(num_workers=8)
tokenizer = SwTokenizer('some_tokenizer.json', kiwi=kiwi)
tokenizer.encode(...) 
# kiwi.num_workers가 8이기 때문에 8개 스레드에서 작업이 처리됨
tokenizer = SwTokenizer('some_tokenizer.json', num_workers=8)
tokenizer.encode(...) 
# 위와 동일하게 8개 스레드 사용
```
        '''
        self.kiwi.space_tolerance = self._space_tolerance
        return super().encode(text, return_offsets)
    
    def encode_from_morphs(self, 
        morphs: Iterable[Union[Tuple[str, str, bool], Tuple[str, str]]],
        return_offsets: bool = False,
    ) -> List[int]:
        '''
이미 형태소 분석이 완료된 결과를 토큰화하여 token id의 리스트로 변환합니다.

Parameters
----------
morphs: Iterable[Union[Tuple[str, str, bool], Tuple[str, str]]]
    `(형태, 품사태그, 왼쪽의 공백 여부)` 혹은 `(형태, 품사태그)`로 구성된 tuple의 리스트
    `왼쪽의 공백 여부`는 생략 가능하며 이 경우 `False`로 처리됩니다.

return_offsets: bool
    True일 경우 각 토큰들의 형태소 상의 시작지점 및 끝지점이 함께 반환됩니다.
    
Returns
-------
token_ids: List[int]
    `return_offsets = False`인 경우.
    token id의 목록을 list로 반환합니다.

token_ids_and_offsets: Tuple[List[int], List[Tuple[int, int]]]
    `return_offsets = True`인 경우.
    token id의 리스트와 각 토큰들의 시작지점과 끝지점(형태소 단위)을 나타내는 tuple의 리스트를 반환합니다.

        '''
        self.kiwi.space_tolerance = self._space_tolerance
        return super().encode_from_morphs(morphs, return_offsets)

    def tokenize_encode(self, 
        text: Union[str, Iterable[str]],
        return_offsets: bool = False,
    ) -> Union[Tuple[List[Token], List[int]], Tuple[List[Token], List[int], List[Tuple[int, int]]]]:
        '''
주어진 텍스트의 형태소 분석 결과 및 토큰화 결과를 함께 반환합니다.

Parameters
----------
text: Union[str, Iterable[str]]
    분할할 텍스트, 혹은 텍스트의 iterable
    이 인자를 단일 str로 줄 경우, 싱글스레드에서 처리하며
    str의 Iterable로 줄 경우, 멀티스레드로 분배하여 처리합니다.

return_offsets: bool
    True일 경우 각 토큰들의 텍스트 상의 시작지점 및 끝지점이 함께 반환됩니다.

Returns
-------
morphs_and_token_ids: Tuple[List[Token], List[int]]
    `return_offsets = False`인 경우.
    형태소 분석 결과와 token id의 목록을 tuple로 반환합니다.

morphs_token_ids_and_offsets: Tuple[List[Token], List[int], List[Tuple[int, int]]]
    `return_offsets = True`인 경우.
    형태소 분석 결과와 token id의 목록, 각 토큰들의 시작지점과 끝지점(형태소 단위)을 나타내는 tuple의 리스트를 반환합니다.

Notes
-----

        '''
        self.kiwi.space_tolerance = self._space_tolerance

        def _refine(res):
            morphs, *etc = res
            return (morphs[0][0], *etc)

        ret = super().tokenize_encode(text, return_offsets)
        if isinstance(text, str):
            return _refine(ret)
        else:
            return map(_refine, ret)

    def decode(self,
        ids: Iterable[int],
        ignore_errors: bool = True,
    ) -> str:
        '''
token id의 리스트를 다시 조합하여 텍스트로 변환합니다.

Parameters
----------
ids: Iterable[int]
    token id의 리스트
ignore_errors: bool
    token을 유니코드 텍스트로 복원할 때 발생하는 오류를 무시할지 설정합니다.
    기본값은 True, 이 경우 오류가 난 부분은 대체문자로 대체됩니다.
    False인 경우 오류가 났을 때 예외를 발생시킵니다.
    
Returns
-------
decoded: str
    다시 조합된 텍스트

Notes
-----
`encode`시 정규화 과정에서 공백 문자나 구두점, 한자 등의 띄어쓰기 변경이 발생하므로
`encode`한 결과를 `decode`한다고 해서 항상 동일한 결과가 나오지는 않습니다.
        '''
        return super().decode(ids, ignore_errors)
    
    def save(self, path:str):
        '''
현재 토크나이저를 json 파일로 저장합니다.
Parameters
----------
path: str
    저장할 파일의 경로

        '''
        return super().save(path)

    @property
    def vocab(self) -> Dict[str, int]:
        '''
토크나이저에 속한 어휘 집합을 dict로 반환합니다. (읽기 전용)
        '''
        try:
            return self._cached_vocab
        except AttributeError:
            self._cached_vocab = super()._vocab
            return self._cached_vocab

    @property
    def id2vocab(self) -> List[str]:
        try:
            return self._cached_id2vocab
        except AttributeError:
            ret = [None] * len(self)
            for k, v in self.vocab.items():
                ret[v] = k
            self._cached_id2vocab = ret
            return self._cached_id2vocab

    @property
    def config(self) -> SwTokenizerConfig:
        '''
토크나이저의 설정값을 담은 `SwTokenizerConfig` 데이터 클래스를 반환합니다. (읽기 전용)
        '''
        try:
            return self._cached_config
        except:
            self._cached_config = SwTokenizerConfig(**super()._config)
            return self._cached_config

    @property
    def kiwi(self) -> Kiwi:
        '''
토크나이저가 사용하는 Kiwi 인스턴스를 반환합니다. (읽기 전용)
        '''
        return super()._kiwi

    def __repr__(self) -> str:
        return super().__repr__()

    @property
    def unk_token(self) -> Optional[str]:
        return self.config.unk_token
    
    @property
    def pad_token(self) -> Optional[str]:
        return self.config.pad_token

    @property
    def mask_token(self) -> Optional[str]:
        return self.config.mask_token

    @property
    def cls_token(self) -> Optional[str]:
        return self.config.cls_token
    
    @property
    def sep_token(self) -> Optional[str]:
        return self.config.sep_token
    
    @property
    def bos_token(self) -> Optional[str]:
        return self.config.bos_token
    
    @property
    def eos_token(self) -> Optional[str]:
        return self.config.eos_token

    @property
    def unk_token_id(self) -> Optional[int]:
        t = self.config.unk_token
        if t is None: return None
        return self.vocab[t]
    
    @property
    def pad_token_id(self) -> Optional[int]:
        t = self.config.pad_token
        if t is None: return None
        return self.vocab[t]

    @property
    def mask_token_id(self) -> Optional[int]:
        t = self.config.mask_token
        if t is None: return None
        return self.vocab[t]

    @property
    def cls_token_id(self) -> Optional[int]:
        t = self.config.cls_token
        if t is None: return None
        return self.vocab[t]
    
    @property
    def sep_token_id(self) -> Optional[int]:
        t = self.config.sep_token
        if t is None: return None
        return self.vocab[t]
    
    @property
    def bos_token_id(self) -> Optional[int]:
        t = self.config.bos_token
        if t is None: return None
        return self.vocab[t]
    
    @property
    def eos_token_id(self) -> Optional[int]:
        t = self.config.eos_token
        if t is None: return None
        return self.vocab[t]

    @property
    def all_special_tokens(self):
        ret = []
        for t in SPECIAL_TOKEN_NAMES:
            v = getattr(self, t + '_token')
            if v is not None:
                ret.append(v)
        return ret

    @property
    def all_special_ids(self):
        ret = []
        for t in SPECIAL_TOKEN_NAMES:
            v = getattr(self, t + '_token_id')
            if v is not None:
                ret.append(v)
        return ret

    @staticmethod
    def train(
        save_path: Union[str, Iterable[str]],
        texts: Iterable[str],
        config: SwTokenizerConfig,
        vocab_size: Union[int, Iterable[int]],
        chr_coverage: float = 0.9995,
        prefix_min_cnt: int = 5,
        prefix_max_length: int = 15,
        strict_reduction: bool = False,
        remove_repetitive: bool = True,
        prevent_mixed_digit_tokens: bool = True,
        iterations: int = 1000,
        reduction_ratio: float = 0.1,
        kiwi: Optional[Kiwi] = None,
        num_workers: Optional[int] = None,
        show_progress: bool = True,
        total_texts: Optional[int] = None,
        callback : Optional[Union[TrainerCallback, List[TrainerCallback]]] = None,
    ) -> 'SwTokenizer':
        '''
주어진 텍스트로부터 유니그램 언어모델과 형태소 분석을 통합한 알고리즘을 통해 서브워드 토크나이저를 학습합니다. 

또한 목표하는 `vocab_size`가 서로 다른 여러 벌의 토크나이저를 한 번에 학습하는 것도 가능합니다.
이 경우 `save_path`를 str의 리스트로, `vocab_size`를 int의 리스트로 입력해야 합니다.

Parameters
----------
save_path: Union[str, Iterable[str]]
    학습된 토크나이저가 저장될 파일 이름을 지정합니다.

texts: Iterable[str]
    학습에 사용될 말뭉치 텍스트를 지정합니다.

config: SwTokenizerConfig
    토크나이저 설정을 지정합니다.

vocab_size: Union[int, Iterable[int]]
    토크나이저의 어휘 집합의 상한치를 지정합니다.

chr_coverage: float
    어휘 집합의 후보를 구성할 때 말뭉치에 등장한 글자 중 최대 얼마까지를 다룰지를 지정합니다.
    기본값은 0.9995로, 이 경우 말뭉치에 등장한 글자들 중 최대 99.95%를 다룰 수 있도록 어휘 집합을 구성합니다.
    이는 한자나 이모지, 일부 특수기호처럼 종류는 많지만 각각의 등장빈도가 낮은 글자를 배제하여 어휘 집합의 크기를 줄이는 데에 유용합니다.

prefix_min_cnt: int
    어휘 집합의 후보를 구성할 때 최소 몇 번 이상 등장한 접두어를 포함할지 지정합니다.
    기본값은 5로, 이 경우 말뭉치에서 최소 5번 이상 등장한 접두어들만이 어휘 집합의 후보로 들어갑니다.

prefix_max_length: int
    어휘 집합의 후보를 구성할 때 포함되는 접두어의 최대 길이를 지정합니다.
    기본값은 15로, 이 경우 최대 15글자로 구성되는 접두어까지 어휘 집합의 후보로 들어갈 수 있습니다.

prevent_mixed_digit_tokens: bool
    어휘 집합의 후보를 구성할 때 숫자와 다른 문자가 섞인 토큰을 배제합니다.
    기본값은 True로, 이 경우 "1분", "10시" 등의 표현은 항상 `1` `분`, `10` `시`처럼 분할됩니다.

strict_reduction: bool
    어휘 집합을 줄여나갈 때 한 번 배제된 후보가 다시는 사용되지 못하도록 엄격하게 설정합니다. 
    기본값은 False입니다.
    True일 경우, 배제된 후보를 다시 고려할 필요가 없으므로 학습 속도가 약간 빨라집니다.
    반면 False일 경우, 학습 속도가 약간 느려지지만 배제된 후보 중 일부를 상황에 따라 다시 후보에 넣을 수 있으므로
    더 좋은 어휘 집합을 찾아낼 가능성이 있습니다.

remove_repetitive: bool
    어휘 집합을 구성할때 특정 패턴이 여러번 반복되는 접두어(ex: 안녕안녕안녕)를 배제합니다.
    기본값은 True입니다.
    
iteration: int
    어휘 집합을 줄여나가는 과정을 최대 몇 번 반복할지 설정합니다.
    기본값은 1000이지만, 대체로 1000회보다 더 작은 횟수 안에 학습이 완료됩니다.

reduction_ratio: float 
    어휘 집합을 줄여나갈 비율을 설정합니다.
    기본값은 0.1으로, 이 경우 1회 반복시마다 어휘 집합 내의 어휘 중 최대 10%를 제거합니다.
    이 값이 클 수록 더 빨리 학습이 완료되지만, 너무 과감하게 어휘들을 제거하므로
    최종 생성된 어휘 집합의 품질이 낮을 수 있습니다.

kiwi: Kiwi
    학습에 사용할 Kiwi 형태소 분석기 인스턴스입니다.
    생략하거나 None으로 줄 경우 내부에서 자체적으로 Kiwi 인스턴스를 생성하여 사용합니다.
    
num_workers: int
    멀티스레딩 시 사용할 스레드의 개수입니다.
    `kiwi`를 생략한 경우에만 유효합니다.
    `kiwi`를 지정한 경우 `kiwi.num_workers` 값에 의해 멀티스레딩을 수행합니다.

show_progress: bool
    True일 경우 학습의 진행상황을 화면에 보여줍니다. 기본값은 True입니다.

total_texts: int
    `texts`의 전체 길이를 지정합니다. 이 값은 오직 `show_progress=True`일때만 유효하며
    전체 입력의 길이를 이용해 현재 학습이 얼마나 진행되었는지 보여주는 데에 쓰입니다.
    생략 시 `len(texts)`을 이용해 입력의 길이를 스스로 추정합니다.

callback: Union[TrainerCallback, List[TrainerCallback]]
    학습 진행상황을 보고 받을 callback 클래스의 인스턴스를 설정합니다.
    각 인스턴스들은 `TrainerCallback` 클래스를 상속 받은 클래스여야 합니다.

Returns
-------
tokenizer: SwTokenizer
    `save_path`와 `vocab_size`가 각각 `str`, `int`로 주어진 경우.
    학습이 완료된 후 얻어진 토크나이저를 반환합니다.
    또한 토크나이저는 `save_path`로 지정된 경로에 저장되어 있으니
    `SwTokenizer(save_path)`를 통해 새로 읽어들이는 것도 가능합니다.

tokenizers: List[SwTokenizer]
    `save_path`와 `vocab_size`가 각각 `Iterable[str]`, `Iterable[int]`로 주어진 경우.
    학습이 완료된 후 얻어진 토크나이저의 리스트를 반환합니다.
    또한 토크나이저는 `save_path`로 지정된 경로 각각에 저장되어 있으니
    `SwTokenizer(save_path[i])`를 통해 새로 읽어들이는 것도 가능합니다.

Notes
-----
```python
```
        '''
        if kiwi is None:
            kiwi = Kiwi(num_workers=num_workers, load_multi_dict=False)
        elif num_workers is not None:
            raise ValueError("You cannot specify `num_workers` value if you give `kiwi` value.")
                
        if not isinstance(kiwi, Kiwi):
            raise ValueError("`kiwi` must be an instance of `Kiwi`.")
        
        if not isinstance(config, SwTokenizerConfig):
            raise ValueError("`config` must be an instance of `SwTokenizerConfig`.")

        if callback is None:
            callback = []
        elif isinstance(callback, TrainerCallback):
            callback = [callback]
        else:
            callback = list(callback)
        
        if not all(isinstance(c, TrainerCallback) for c in callback):
            raise ValueError("`callback` must be an instance of `TrainerCallback`.")

        if show_progress:
            if total_texts is None:
                try: total_texts = len(texts)
                except: pass
            callback.insert(0, _ProgressShower(None if show_progress is True else show_progress, total_texts, iterations))

        single_target = False
        if isinstance(save_path, str):
            if not isinstance(vocab_size, int):
                raise ValueError("`save_path` should have the same number of elements to `vocab_size`.")
            single_target = True
            save_path = [save_path]
            vocab_size = [vocab_size]
        
        _SwTokenizer._train(
            save_path, 
            texts, 
            config, 
            vocab_size, 
            iterations,
            prefix_min_cnt,
            prefix_max_length,
            strict_reduction, 
            remove_repetitive, 
            prevent_mixed_digit_tokens,
            chr_coverage, 
            reduction_ratio, 
            kiwi, 
            callback,
        )
        if single_target:
            return SwTokenizer(save_path[0], kiwi)
        else:
            return [SwTokenizer(s, kiwi) for s in save_path]




'''
.. versionadded:: 0.15.1

`sw_trainer` 모듈은 간편하게 SwTokenizer를 학습할 수 있는 CLI 환경을 제공합니다.

간단 예시:
```bash
python3 -m kiwipiepy.sw_trainer \\
    some_corpus.txt \\
    --save_path my_tokenizer.json \\
    --vocab_size 32000
```
`some_corpus.txt` 말뭉치 파일을 분석하여 `32000`종류의 vocab을 가지는 SwTokenizer를 학습해 `my_tokenizer.json`으로 저장합니다.

옵션으로 줄 수 있는 인자는 아래와 같습니다.
```text
--lowercase
    True or False (기본값: False)
    토큰화에 앞서 대문자들을 전부 소문자로 정규화합니다.

--split_chinese
    True or False (기본값: True)
    토큰화에 앞서 모든 한자를 한 글자씩 분리합니다.

--whole_word_unk
    True or False (기본값: False)
    어휘 집합에 속하지 않은 글자가 
    하나라도 포함된 어절 전체를 UNK토큰으로 지정합니다.

--split_punct
    True or False (기본값: True)
    토큰화에 앞서 구두점을 분리합니다.

--simple_tag
    True or False (기본값: True)
    세부 품사태그(예: VV, VX, VA) 대신 대표 품사태그(예: V)를 사용합니다.

--split_verb
    True or False (기본값: True)
    서브워드로 분리가능한 동사가 있다면 더 작은 단위로 분리합니다. 
    (예: 뛰어가다 -> 뛰/V 어/E 가/V 다/E)

--split_eomi
    True or False (기본값: True)
    서브워드로 분리가능한 어미가 있다면 더 작은 단위로 분리합니다. 
    (예: 어요 -> 어/E 요/J)

--use_glue_token
    True or False (기본값: True)
    단어를 이어붙이는 특수 토큰인 Glue토큰을 사용합니다.

--fallback_hangul
    True or False (기본값: True)
    토크나이저의 어휘 집합에 속하지 않은 한글이 있다면 
    초성+중성 / 종성의 형태로 분해하여 처리합니다.

--fallback_byte
    True or False (기본값: False)
    토크나이저의 어휘 집합에 속하지 않은 문자가 있다면 
    UTF8 Byte로 변환하여 처리합니다.

--use_newline_token
    True or False (기본값: False)
    줄바꿈 문자를 토큰으로 사용합니다. False일 경우 공백으로 취급하여 무시합니다.
    줄바꿈 문자로는 fallback_byte 토큰의 10번째 바이트 토큰(\\n)을 사용합니다.

--unk_token
--cls_token
--sep_token
--pad_token
--mask_token
--bos_token
--eos_token
    str
    특수 토큰을 지정합니다.

--save_path
    str, 여러 개 지정 가능
    학습된 SwTokenizer의 저장 경로를 설정합니다.
    vocab_size를 여러 개 지정한 경우 각각의 vocab_size별로 save_path를 지정해주어야 합니다.

--vocab_size
    int, 여러 개 지정 가능
    토크나이저의 어휘 집합의 상한치를 지정합니다.

--chr_coverage
    float (기본값: 0.9995)
    어휘 집합의 후보를 구성할 때 말뭉치에 등장한 글자 중 최대 얼마까지를 
    다룰지를 지정합니다. 기본값은 0.9995로, 이 경우 말뭉치에 등장한 글자들 중
    최대 99.95%를 다룰 수 있도록 어휘 집합을 구성합니다. 이는 한자나 이모지, 
    일부 특수기호처럼 종류는 많지만 각각의 등장빈도가 낮은 글자를 배제하여 
    어휘 집합의 크기를 줄이는 데에 유용합니다.

--prefix_min_cnt
    int (기본값: 5)
    어휘 집합의 후보를 구성할 때 최소 몇 번 이상 등장한 접두어를 포함할지 지정합니다.

--prefix_max_length
    int (기본값: 15)
    어휘 집합의 후보를 구성할 때 포함되는 접두어의 최대 길이를 지정합니다.

--prevent_mixed_digit_tokens
    True or False (기본값: True)
    어휘 집합의 후보를 구성할 때 숫자와 다른 문자가 섞인 토큰을 배제합니다.

--strict_reduction
    True or False (기본값: False)
    어휘 집합을 줄여나갈 때 한 번 배제된 후보가 다시는 사용되지 못하도록 엄격하게 설정합니다. 

--remove_repetitive
    True or False (기본값: True)
    어휘 집합을 구성할때 특정 패턴이 여러번 반복되는 접두어(ex: 안녕안녕안녕)를 배제합니다.

--iterations
    int (기본값: iteration)
    어휘 집합을 줄여나가는 과정을 최대 몇 번 반복할지 설정합니다.
    기본값은 1000이지만, 대체로 1000회보다 더 작은 횟수 안에 학습이 완료됩니다.

--reduction_ratio
    float (기본값: 0.1)
    어휘 집합을 줄여나갈 비율을 설정합니다.

--num_workers
    int (기본값: 0)
    멀티스레딩 시 사용할 스레드의 개수입니다.
    0일 경우 현재 시스템의 가용한 스레드를 전부 사용합니다.

```
'''

from kiwipiepy.sw_tokenizer import SwTokenizer, SwTokenizerConfig

class MultipleFileLoader:

    def _count_lines(self, path, chunk_size=65536):
        lines = 0
        with open(path, 'rb') as f:
            while True:
                b = f.read(chunk_size)
                if not b: break
                lines += b.count(b'\n')
        return lines

    def __init__(self, pathes):
        self._pathes = list(pathes)
        self._total_lines = sum(map(self._count_lines, self._pathes))
    
    def __iter__(self):
        for p in self._pathes:
            yield from open(p, encoding='utf-8')
    
    def __len__(self):
        return self._total_lines

def main(args):
    config = SwTokenizerConfig(
        lowercase=args.lowercase,
        split_chinese=args.split_chinese,
        whole_word_unk=args.whole_word_unk,
        split_punct=args.split_punct,
        simple_tag=args.simple_tag,
        split_verb=args.split_verb,
        split_eomi=args.split_eomi,
        use_glue_token=args.use_glue_token,
        fallback_hangul=args.fallback_hangul,
        fallback_byte=args.fallback_byte,

        unk_token=args.unk_token,
        cls_token=args.cls_token,
        sep_token=args.sep_token,
        pad_token=args.pad_token,
        mask_token=args.mask_token,
        bos_token=args.bos_token,
        eos_token=args.eos_token,
    )
    print("Building SwTokenizer...")
    p = dict(
        vocab_size=args.vocab_size,
        chr_coverage=args.chr_coverage,
        prefix_min_cnt=args.prefix_min_cnt,
        prefix_max_length=args.prefix_max_length,
        prevent_mixed_digit_tokens=args.prevent_mixed_digit_tokens,
        strict_reduction=args.strict_reduction,
        remove_repetitive=args.remove_repetitive,
        iterations=args.iterations,
        reduction_ratio=args.reduction_ratio,
        num_workers=args.num_workers,
        **config.__dict__
    )
    for k, v in p.items():
        print(f"|  {k}: {v!r}")
    print(flush=True)
    SwTokenizer.train(
        args.save_path,
        MultipleFileLoader(args.input_files),
        config=config,
        vocab_size=args.vocab_size,
        chr_coverage=args.chr_coverage,
        prefix_min_cnt=args.prefix_min_cnt,
        prefix_max_length=args.prefix_max_length,
        prevent_mixed_digit_tokens=args.prevent_mixed_digit_tokens,
        strict_reduction=args.strict_reduction,
        remove_repetitive=args.remove_repetitive,
        iterations=args.iterations,
        reduction_ratio=args.reduction_ratio,
        num_workers=args.num_workers,
    )
    print(f'Tokenizer was saved at {args.save_path}')

if __name__ == '__main__':
    def _bool(v):
        vl = v.lower()
        if vl in ('true', 't', '1'):
            return True
        elif vl in ('false', 'f', '0'):
            return False
        else:
            raise ValueError(f"Wrong value {v} for bool type argument.")

    import argparse
    parser = argparse.ArgumentParser(
        'python -m kiwpiepy.sw_trainer', 
        description='Kiwi SwTokenizer Trainer'
    )
    parser.add_argument('input_files', nargs='+')
    parser.add_argument('--lowercase', default=False, type=_bool)
    parser.add_argument('--split_chinese', default=True, type=_bool)
    parser.add_argument('--whole_word_unk', default=False, type=_bool)
    parser.add_argument('--split_punct', default=True, type=_bool)
    parser.add_argument('--simple_tag', default=True, type=_bool)
    parser.add_argument('--split_verb', default=True, type=_bool)
    parser.add_argument('--split_eomi', default=True, type=_bool)
    parser.add_argument('--use_glue_token', default=True, type=_bool)
    parser.add_argument('--fallback_hangul', default=True, type=_bool)
    parser.add_argument('--fallback_byte', default=False, type=_bool)
    parser.add_argument('--use_newline_token', default=False, type=_bool)
    parser.add_argument('--unk_token', default="[UNK]")
    parser.add_argument('--cls_token')
    parser.add_argument('--sep_token')
    parser.add_argument('--pad_token')
    parser.add_argument('--mask_token')
    parser.add_argument('--bos_token')
    parser.add_argument('--eos_token')
    
    parser.add_argument('--save_path', required=True, nargs='+')
    parser.add_argument('--vocab_size', required=True, type=int, nargs='+')
    parser.add_argument('--chr_coverage', default=0.9995, type=float)
    parser.add_argument('--prefix_min_cnt', default=5, type=int)
    parser.add_argument('--prefix_max_length', default=15, type=int)
    parser.add_argument('--prevent_mixed_digit_tokens', default=True, type=_bool)
    parser.add_argument('--strict_reduction', default=False, type=_bool)
    parser.add_argument('--remove_repetitive', default=True, type=_bool)
    parser.add_argument('--iterations', default=1000, type=int)
    parser.add_argument('--reduction_ratio', default=0.1, type=float)
    parser.add_argument('--num_workers', default=0, type=int)
    
    main(parser.parse_args())



from kiwipiepy._wrap import TypoTransformer, TypoDefinition

_basic_defs = [
    TypoDefinition(["ㅐ", "ㅔ"], ["ㅐ", "ㅔ"], 1.),
    TypoDefinition(["ㅐ", "ㅔ"], ["ㅒ", "ㅖ"], 1.5),
    TypoDefinition(["ㅒ", "ㅖ"], ["ㅐ", "ㅔ"], 1.5),
    TypoDefinition(["ㅒ", "ㅖ"], ["ㅒ", "ㅖ"], 1.),
    TypoDefinition(["ㅚ", "ㅙ", "ㅞ"], ["ㅚ", "ㅙ", "ㅞ", "ㅐ", "ㅔ"], 1.),
    TypoDefinition(["ㅝ"], ["ㅗ", "ㅓ"], 1.),
    TypoDefinition(["ㅟ", "ㅢ"], ["ㅣ"], 1.),
    TypoDefinition(["위", "의"], ["이"], float("inf")),
    TypoDefinition(["위", "의"], ["이"], 1., "any"),
    TypoDefinition(["자", "쟈"], ["자", "쟈"], 1.),
    TypoDefinition(["재", "쟤"], ["재", "쟤"], 1.),
    TypoDefinition(["저", "져"], ["저", "져"], 1.),
    TypoDefinition(["제", "졔"], ["제", "졔"], 1.),
    TypoDefinition(["조", "죠", "줘"], ["조", "죠", "줘"], 1.),
    TypoDefinition(["주", "쥬"], ["주", "쥬"], 1.),
    TypoDefinition(["차", "챠"], ["차", "챠"], 1.),
    TypoDefinition(["채", "챼"], ["채", "챼"], 1.),
    TypoDefinition(["처", "쳐"], ["처", "쳐"], 1.),
    TypoDefinition(["체", "쳬"], ["체", "쳬"], 1.),
    TypoDefinition(["초", "쵸", "춰"], ["초", "쵸", "춰"], 1.),
    TypoDefinition(["추", "츄"], ["추", "츄"], 1.),
    TypoDefinition(["유", "류"], ["유", "류"], 1.),
    TypoDefinition(["므", "무"], ["므", "무"], 1.),
    TypoDefinition(["브", "부"], ["브", "부"], 1.),
    TypoDefinition(["프", "푸"], ["프", "푸"], 1.),
    TypoDefinition(["르", "루"], ["르", "루"], 1.),
    TypoDefinition(["러", "뤄"], ["러", "뤄"], 1.),
    TypoDefinition(["\ㄲ", "\ㄳ"], ["\ㄱ", "\ㄲ", "\ㄳ"], 1.5),
    TypoDefinition(["\ㄵ", "\ㄶ"], ["\ㄴ", "\ㄵ", "\ㄶ"], 1.5),
    TypoDefinition(["\ㄺ", "\ㄻ", "\ㄼ", "\ㄽ", "\ㄾ", "\ㄿ", "\ㅀ"], ["\ㄹ", "\ㄺ", "\ㄻ", "\ㄼ", "\ㄽ", "\ㄾ", "\ㄿ", "\ㅀ"], 1.5),
    TypoDefinition(["\ㅅ", "\ㅆ"], ["\ㅅ", "\ㅆ"], 1.),

    TypoDefinition(["안"], ["않"], 1.5),
    TypoDefinition(["맞추", "맞히"], ["맞추", "맞히"], 1.5),
    TypoDefinition(["맞춰", "맞혀"], ["맞춰", "맞혀"], 1.5),
    TypoDefinition(["받치", "바치", "받히"], ["받치", "바치", "받히"], 1.5),
    TypoDefinition(["받쳐", "바쳐", "받혀"], ["받쳐", "바쳐", "받혀"], 1.5),
    TypoDefinition(["던", "든"], ["던", "든"], 1.),
    TypoDefinition(["때", "데"], ["때", "데"], 1.5),
    TypoDefinition(["빛", "빚"], ["빛", "빚"], 1.),

    TypoDefinition(["\ㄷ이", "지"], ["\ㄷ이", "지"], 1.),
    TypoDefinition(["\ㄷ여", "져"], ["\ㄷ여", "져"], 1.),
    TypoDefinition(["\ㅌ이", "치"], ["\ㅌ이", "치"], 1.),
    TypoDefinition(["\ㅌ여", "쳐"], ["\ㅌ여", "쳐"], 1.),
	
    TypoDefinition(["ㄱ", "ㄲ"], ["ㄱ", "ㄲ"], 1., "applosive"),
    TypoDefinition(["ㄷ", "ㄸ"], ["ㄷ", "ㄸ"], 1., "applosive"),
    TypoDefinition(["ㅂ", "ㅃ"], ["ㅂ", "ㅃ"], 1., "applosive"),
    TypoDefinition(["ㅅ", "ㅆ"], ["ㅅ", "ㅆ"], 1., "applosive"),
    TypoDefinition(["ㅈ", "ㅉ"], ["ㅈ", "ㅉ"], 1., "applosive"),

    TypoDefinition(["\ㅎㅎ", "\ㄱㅎ", "\ㅎㄱ"], ["\ㅎㅎ", "\ㄱㅎ", "\ㅎㄱ"], 1.),

    TypoDefinition(["\ㄱㄴ", "\ㄲㄴ", "ᆪㄴ", "ᆿㄴ", "ᆼㄴ"], ["\ㄱㄴ", "\ㄲㄴ", "ᆪㄴ", "ᆿㄴ", "ᆼㄴ"], 1.),
    TypoDefinition(["\ㄱㅁ", "\ㄲㅁ", "ᆪㅁ", "ᆿㅁ", "ᆼㅁ"], ["\ㄱㅁ", "\ㄲㅁ", "ᆪㅁ", "ᆿㅁ", "ᆼㅁ"], 1.),
    TypoDefinition(["\ㄱㄹ", "\ㄲㄹ", "ᆪㄹ", "ᆿㄹ", "ᆼㄹ", "ᆼㄴ",], ["\ㄱㄹ", "\ㄲㄹ", "ᆪㄹ", "ᆿㄹ", "ᆼㄹ", "ᆼㄴ",], 1.),
    TypoDefinition(["\ㄷㄴ", "\ㅅㄴ", "\ㅆㄴ", "\ㅈㄴ", "ᆾㄴ", "ᇀㄴ", "\ㄴㄴ"], ["\ㄷㄴ", "\ㅅㄴ", "\ㅆㄴ", "\ㅈㄴ", "ᆾㄴ", "ᇀㄴ", "\ㄴㄴ"], 1.),
    TypoDefinition(["\ㄷㅁ", "\ㅅㅁ", "\ㅆㅁ", "\ㅈㅁ", "ᆾㅁ", "ᇀㅁ", "\ㄴㅁ"], ["\ㄷㅁ", "\ㅅㅁ", "\ㅆㅁ", "\ㅈㅁ", "ᆾㅁ", "ᇀㅁ", "\ㄴㅁ"], 1.),
    TypoDefinition(["\ㄷㄹ", "\ㅅㄹ", "\ㅆㄹ", "\ㅈㄹ", "ᆾㄹ", "ᇀㄹ", "\ㄴㄹ", "\ㄴㄴ",], ["\ㄷㄹ", "\ㅅㄹ", "\ㅆㄹ", "\ㅈㄹ", "ᆾㄹ", "ᇀㄹ", "\ㄴㄹ", "\ㄴㄴ",], 1.),
    TypoDefinition(["\ㅂㄴ", "ᆹㄴ", "ᇁㄴ", "\ㅁㄴ"], ["\ㅂㄴ", "ᆹㄴ", "ᇁㄴ", "\ㅁㄴ"], 1.),
    TypoDefinition(["\ㅂㅁ", "ᆹㅁ", "ᇁㅁ", "\ㅁㅁ"], ["\ㅂㅁ", "ᆹㅁ", "ᇁㅁ", "\ㅁㅁ"], 1.),
    TypoDefinition(["\ㅂㄹ", "ᆹㄹ", "ᇁㄹ", "\ㅁㄹ", "\ㅁㄴ",], ["\ㅂㄹ", "ᆹㄹ", "ᇁㄹ", "\ㅁㄹ", "\ㅁㄴ",], 1.),
    TypoDefinition(["\ㄴㄹ", "\ㄴㄴ", "\ㄹㄹ", "\ㄹㄴ"], ["\ㄴㄹ", "\ㄴㄴ", "\ㄹㄹ", "\ㄹㄴ"], 1.),
	
    TypoDefinition(["\ㄱㅇ", "ㄱ"], ["\ㄱㅇ", "ㄱ"], 1., "vowel"),
    TypoDefinition(["\ㄲㅇ", "ㄲ"], ["\ㄲㅇ", "ㄲ"], 1., "vowel"),
    TypoDefinition(["\ㄴㅇ", "\ㄴㅎ", "ㄴ"], ["\ㄴㅇ", "\ㄴㅎ", "ㄴ"], 1., "vowel"),
    TypoDefinition(["\ㄵㅇ", "\ㄴㅈ"], ["\ㄵㅇ", "\ㄴㅈ"], 1., "vowel"),
    TypoDefinition(["\ㄶㅇ", "ㄴ"], ["\ㄶㅇ", "ㄴ"], 1., "vowel"),
    TypoDefinition(["\ㄷㅇ", "ㄷ"], ["\ㄷㅇ", "ㄷ"], 1., "vowel"),
    TypoDefinition(["\ㄹㅇ", "\ㄹㅎ", "ㄹ"], ["\ㄹㅇ", "\ㄹㅎ", "ㄹ"], 1., "vowel"),
    TypoDefinition(["\ㄺㅇ", "\ㄹㄱ"], ["\ㄺㅇ", "\ㄹㄱ"], 1., "vowel"),
    TypoDefinition(["\ㄺㅎ", "\ㄹㅋ"], ["\ㄺㅎ", "\ㄹㅋ"], 1., "vowel"),
    TypoDefinition(["\ㅁㅇ", "ㅁ"], ["\ㅁㅇ", "ㅁ"], 1., "vowel"),
    TypoDefinition(["\ㅂㅇ", "ㅂ"], ["\ㅂㅇ", "ㅂ"], 1., "vowel"),
    TypoDefinition(["\ㅅㅇ", "ㅅ"], ["\ㅅㅇ", "ㅅ"], 1., "vowel"),
    TypoDefinition(["\ㅆㅇ", "\ㅅㅅ", "ㅆ"], ["\ㅆㅇ", "\ㅅㅅ", "ㅆ"], 1., "vowel"),
    TypoDefinition(["\ㅈㅇ", "ㅈ"], ["\ㅈㅇ", "ㅈ"], 1., "vowel"),
    TypoDefinition(["\ㅊㅇ", "\ㅊㅎ", "\ㅈㅎ", "ㅊ"], ["\ㅊㅇ", "\ㅊㅎ", "\ㅈㅎ", "ㅊ"], 1., "vowel"),
    TypoDefinition(["\ㅋㅇ", "\ㅋㅎ", "\ㄱㅎ", "ㅋ"], ["\ㅋㅇ", "\ㅋㅎ", "\ㄱㅎ", "ㅋ"], 1., "vowel"),
    TypoDefinition(["\ㅌㅇ", "\ㅌㅎ", "\ㄷㅎ", "ㅌ"], ["\ㅌㅇ", "\ㅌㅎ", "\ㄷㅎ", "ㅌ"], 1., "vowel"),
    TypoDefinition(["\ㅍㅇ", "\ㅍㅎ", "\ㅂㅎ", "ㅍ"], ["\ㅍㅇ", "\ㅍㅎ", "\ㅂㅎ", "ㅍ"], 1., "vowel"),

    TypoDefinition(["은", "는"], ["은", "는"], 2.),
    TypoDefinition(["을", "를"], ["을", "를"], 2.),

    TypoDefinition(["ㅣ워", "ㅣ어", "ㅕ"], ["ㅣ워", "ㅣ어", "ㅕ"], 1.5),
    #TypoDefinition(["ㅡ아", "ㅏ"], ["ㅡ아", "ㅏ"], 1.5),
    #TypoDefinition(["ㅡ어", "ㅓ"], ["ㅡ어", "ㅓ"], 1.5),
    #TypoDefinition(["ㅡ오", "ㅗ"], ["ㅡ오", "ㅗ"], 1.5),
    #TypoDefinition(["ㅡ우", "ㅜ"], ["ㅡ우", "ㅜ"], 1.5),
]


_continual_defs = [
    TypoDefinition(["\ㄳ"], ["\ㄱ\ㅅ", "\ㄱ\ㅆ"], 1e-12),
    TypoDefinition(["\ㄵ"], ["\ㄴ\ㅈ"], 1e-12),
    TypoDefinition(["\ㄶ"], ["\ㄴ\ㅎ"], 1e-12),
    TypoDefinition(["\ㄺ"], ["\ㄹ\ㄱ"], 1e-12),
    TypoDefinition(["\ㄻ"], ["\ㄹ\ㅁ"], 1e-12),
    TypoDefinition(["\ㄼ"], ["\ㄹ\ㅂ"], 1e-12),
    TypoDefinition(["\ㄽ"], ["\ㄹ\ㅅ"], 1e-12),
    TypoDefinition(["\ㄾ"], ["\ㄹ\ㅌ"], 1e-12),
    TypoDefinition(["\ㄿ"], ["\ㄹ\ㅍ"], 1e-12),
    TypoDefinition(["\ㅀ"], ["\ㄹ\ㅎ"], 1e-12),
    TypoDefinition(["\ㅄ"], ["\ㅂ\ㅅ", "\ㅂ\ㅆ"], 1e-12),
]

basic_typos = TypoTransformer(_basic_defs)
'''
내장된 기본 오타 자료입니다.
자주 틀리는 오타들을 포함하고 있습니다. 형태소 내의 오타만 교정할 수 있습니다.
'''

continual_typos = TypoTransformer(_continual_defs, continual_typo_cost=1.)
'''
내장된 기본 연철 오타 자료입니다.
받침 + 초성 ㅇ/ㅎ을 잘못 이어적은 경우(먹어라 -> 머거라, 않지 -> 안치 등)를 교정할 수 있습니다.
형태소 간의 오타만 교정할 수 있습니다. 
'''

basic_typos_with_continual = basic_typos | continual_typos
'''
기본 오타와 연철 오타를 합친 것입니다.
형태소 내의 오타와 형태소 간의 오타를 모두 교정할 수 있습니다.
'''

lengthening_typos = TypoTransformer([], lengthening_typo_cost=0.25)
'''
장음화 오타 자료입니다. 음절 하나를 두 음절 이상으로 길게 늘여적은 경우(진짜 -> 지인짜)를 교정할 수 있습니다.
'''

basic_typos_with_continual_and_lengthening = basic_typos_with_continual | lengthening_typos
'''
기본 오타, 연철 오타, 장음화 오타를 합친 것입니다.
형태소 내의 오타, 형태소 간의 오타, 장음화 오타를 모두 교정할 수 있습니다.
'''



'''
const 모듈은 kiwipiepy에서 사용되는 주요 상수값들을 모아놓은 모듈입니다.
'''
from enum import IntFlag

class Match(IntFlag):
    """
    .. versionadded:: 0.8.0

    분석 시 특수한 문자열 패턴 중 어떤 것들을 추출할 지 선택할 수 있습니다.
    bitwise OR 연산으로 여러 개 선택하여 사용가능합니다.
    """
    
    URL = 1 << 0
    """ 인터넷 주소 형태의 텍스트를 W_URL이라는 태그로 추출합니다. """
    
    EMAIL = 1 << 1
    """ 이메일 주소 형태의 텍스트를 W_EMAIL이라는 태그로 추출합니다. """
    
    HASHTAG = 1 << 2
    """ 해시태그(#해시태그) 형태의 텍스트를 W_HASHTAG라는 태그로 추출합니다. """
    
    MENTION = 1 << 3
    """
    멘션(@멘션) 형태의 텍스트를 W_MENTION이라는 태그로 추출합니다.
    
    .. versionadded:: 0.8.2
    """
    
    SERIAL = 1 << 4
    """
    일련번호 형태의 텍스트를 W_SERIAL이라는 태그로 추출합니다.
    
    .. versionadded:: 0.14.0
    """
    
    EMOJI = 1 << 5
    """
    이모지 형태의 텍스트를 W_EMOJI라는 태그로 추출합니다.

    .. versionadded:: 0.18.0
    """
    
    ALL = URL | EMAIL | HASHTAG | MENTION | SERIAL | EMOJI
    """ URL, EMAIL, HASHTAG, MENTION, SERIAL, EMOJI을 모두 사용합니다. """
    
    NORMALIZING_CODA = 1 << 16
    """ '먹었엌ㅋㅋ'처럼 받침이 덧붙어서 분석에 실패하는 경우, 받침을 분리하여 정규화합니다. """
    
    JOIN_NOUN_PREFIX = 1 << 17
    """
    명사의 접두사를 분리하지 않고 결합합니다. 풋/XPN 사과/NNG -> 풋사과/NNG 

    .. versionadded:: 0.11.0
    """
    
    JOIN_NOUN_SUFFIX = 1 << 18
    """
    명사의 접미사를 분리하지 않고 결합합니다. 사과/NNG 들/XSN -> 사과들/NNG
    
    .. versionadded:: 0.11.0
    """
    
    JOIN_VERB_SUFFIX = 1 << 19
    """
    동사 파생접미사를 분리하지 않고 결합합니다. 사랑/NNG 하/XSV 다/EF -> 사랑하/VV 다/EF

    .. versionadded:: 0.11.0
    """
    
    JOIN_ADJ_SUFFIX = 1 << 20
    """
    형용사 파생접미사를 분리하지 않고 결합합니다. 매콤/XR 하/XSA 다/EF -> 매콤하/VA 다/EF

    .. versionadded:: 0.11.0
    """
    
    JOIN_ADV_SUFFIX = 1 << 21
    """
    부사 파생접미사를 분리하지 않고 결합합니다. 요란/XR 히/XSM -> 요란히/MAG

    .. versionadded:: 0.15.0
    """
    
    SPLIT_COMPLEX = 1 << 22
    """
    더 잘게 분할 가능한 형태소를 모두 분할합니다. 고마움/NNG -> 고맙/VA-I 음/ETN

    .. versionadded:: 0.15.0
    """
    
    Z_CODA = 1 << 23
    """
    조사/어미에 덧붙은 받침을 Z_CODA 태그로 분리합니다. 했어욗 -> 하/VV 었/EP 어요/EF ㄳ/Z_CODA

    .. versionadded:: 0.15.0
    """
    
    COMPATIBLE_JAMO = 1 << 24
    """
    형태소 분석 결과 출력 시 첫가끝 자모를 호환용 자모로 변환합니다.

    .. versionadded:: 0.18.1
    """
    
    SPLIT_SAISIOT = 1 << 25
    """
    사이시옷이 포함된 합성명사를 분리합니다. 만둣국 -> 만두/NNG ᆺ/Z_SIOT 국/NNG
    
    .. versionadded:: 0.20.0
    """
    
    MERGE_SAISIOT = 1 << 26
    """
    사이시옷이 포함된 것으로 추정되는 명사를 결합합니다. 만둣국 -> 만둣국/NNG

    .. versionadded:: 0.20.0
    """
    
    JOIN_V_SUFFIX = JOIN_VERB_SUFFIX | JOIN_ADJ_SUFFIX
    """
    동사/형용사형 파생접미사를 분리하지 않고 결합합니다.

    .. versionadded:: 0.11.0
    """
    
    JOIN_AFFIX = JOIN_NOUN_PREFIX | JOIN_NOUN_SUFFIX | JOIN_V_SUFFIX | JOIN_ADV_SUFFIX
    """
    모든 접두사/접미사를 분리하지 않고 결합합니다.

    .. versionadded:: 0.11.0
    """

class Dialect(IntFlag):
    """
    .. versionadded:: 0.22.0

    방언 정보를 나타내는 열거형입니다.
    """
    
    STANDARD = 0
    """ 표준어 """
    표준 = STANDARD
    
    GYEONGGI = 1 << 0
    """ 경기 방언 """
    경기 = GYEONGGI

    CHUNGCHEONG = 1 << 1
    """ 충청 방언 """
    충청 = CHUNGCHEONG

    GANGWON = 1 << 2
    """ 강원 방언 """
    강원 = GANGWON

    GYEONGSANG = 1 << 3
    """ 경상 방언 """
    경상 = GYEONGSANG

    JEOLLA = 1 << 4
    """ 전라 방언 """
    전라 = JEOLLA

    JEJU = 1 << 5
    """ 제주 방언 """
    제주 = JEJU

    HWANGHAE = 1 << 6
    """ 황해 방언 """
    황해 = HWANGHAE

    HAMGYEONG = 1 << 7
    """ 함경 방언 """
    함경 = HAMGYEONG

    PYEONGAN = 1 << 8
    """ 평안 방언 """
    평안 = PYEONGAN

    ARCHAIC = 1 << 9
    """ 옛말 """
    옛말 = ARCHAIC

    ALL = (GYEONGGI | CHUNGCHEONG | GANGWON | GYEONGSANG | JEOLLA | JEJU | HWANGHAE | HAMGYEONG | PYEONGAN | ARCHAIC)
    """ 모든 방언 """



import tempfile

from transformers import AutoTokenizer
import kiwipiepy.transformers_addon

def test_init():
    global tokenizer
    tokenizer = AutoTokenizer.from_pretrained('test/sample_tokenizer')

def test_basic_tokenize():
    e = tokenizer("가자")
    assert (e['input_ids'] == [2, 75, 130, 3])
    assert (e['attention_mask'] == [1, 1, 1, 1])
    assert (e['token_type_ids'] == [0, 0, 0, 0])

    e = tokenizer("가자", "맞습니다요")
    assert (e['input_ids'] == [2, 75, 130, 3, 282, 64, 157, 3])
    assert (e['attention_mask'] == [1, 1, 1, 1, 1, 1, 1, 1])
    assert (e['token_type_ids'] == [0, 0, 0, 0, 1, 1, 1, 1])

    e = tokenizer(["가자"])
    assert (e['input_ids'] == [[2, 75, 130, 3]])
    assert (e['attention_mask'] == [[1, 1, 1, 1]])
    assert (e['token_type_ids'] == [[0, 0, 0, 0]])
    
    e = tokenizer([("가자", "맞습니다요")])
    assert (e['input_ids'] == [[2, 75, 130, 3, 282, 64, 157, 3]])
    assert (e['attention_mask'] == [[1, 1, 1, 1, 1, 1, 1, 1]])
    assert (e['token_type_ids'] == [[0, 0, 0, 0, 1, 1, 1, 1]])

def test_without_special_tokens():
    e = tokenizer("가자", add_special_tokens=False)
    assert (e['input_ids'] == [75, 130])
    assert (e['attention_mask'] == [1, 1])
    assert (e['token_type_ids'] == [0, 0])

    e = tokenizer("가자", "맞습니다요", add_special_tokens=False)
    assert (e['input_ids'] == [75, 130, 282, 64, 157])
    assert (e['attention_mask'] == [1, 1, 1, 1, 1])
    assert (e['token_type_ids'] == [0, 0, 1, 1, 1])

def test_truncation():
    e = tokenizer("가자가자가자~", "맞습니다요맞구요맞았어요!", truncation=True, max_length=14)
    assert (e['input_ids'] == [2, 75, 130, 75, 130, 75, 3, 282, 64, 157, 282, 1219, 282, 3])
    assert (e['attention_mask'] == [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1])
    assert (e['token_type_ids'] == [0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1])

    e = tokenizer("가자가자가자~", "맞습니다요맞구요맞았어요!", truncation='only_first', max_length=14)
    assert (e['input_ids'] == [2, 75, 130, 3, 282, 64, 157, 282, 1219, 282, 11, 78, 85, 3])
    assert (e['attention_mask'] == [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1])
    assert (e['token_type_ids'] == [0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1])

    e = tokenizer("가자가자가자~", "맞습니다요맞구요맞았어요!", truncation='only_second', max_length=14)
    assert (e['input_ids'] == [2, 75, 130, 75, 130, 75, 130, 60, 3, 282, 64, 157, 282, 3])
    assert (e['attention_mask'] == [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1])
    assert (e['token_type_ids'] == [0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1])

def test_pad():
    e = tokenizer("가자", padding='max_length', max_length=8)
    assert (e['input_ids'] == [2, 75, 130, 3, 0, 0, 0, 0])
    assert (e['attention_mask'] == [1, 1, 1, 1, 0, 0, 0, 0])
    assert (e['token_type_ids'] == [0, 0, 0, 0, 0, 0, 0, 0])

    e = tokenizer(["가자", "맞습니다요"], padding=True)
    assert (e['input_ids'] == [[2, 75, 130, 3, 0], [2, 282, 64, 157, 3]])
    assert (e['attention_mask'] == [[1, 1, 1, 1, 0], [1, 1, 1, 1, 1]])
    assert (e['token_type_ids'] == [[0, 0, 0, 0, 0], [0, 0, 0, 0, 0]])

def test_offset_mapping():
    e = tokenizer("맞습니다요!", padding='max_length', max_length=8, return_offsets_mapping=True)
    assert (e['input_ids'] == [2, 282, 27336, 85, 3, 0, 0, 0])
    assert (e['offset_mapping'] == [(0, 0), (0, 1), (1, 5), (5, 6), (0, 0), (0, 0), (0, 0), (0, 0)])

    e = tokenizer("가자", "맞습니다요", return_offsets_mapping=True)
    assert (e['input_ids'] == [2, 75, 130, 3, 282, 64, 157, 3])
    assert (e['offset_mapping'] == [(0, 0), (0, 1), (1, 2), 
                                    (0, 0), (0, 1), (1, 4), (4, 5), (0, 0)])

    e = tokenizer("가자가자가자~", "맞습니다요맞구요맞았어요!", truncation=True, max_length=14, return_offsets_mapping=True)
    assert (e['input_ids'] == [2, 75, 130, 75, 130, 75, 3, 282, 64, 157, 282, 1219, 282, 3])
    assert (e['offset_mapping'] == [(0, 0), (0, 1), (1, 2), (2, 3), (3, 4), (4, 5), 
                                    (0, 0), (0, 1), (1, 4), (4, 5), (5, 6), (6, 8), (8, 9), (0, 0)])

def test_decode():
    d = tokenizer.decode([2, 75, 130, 3])
    assert d == "[CLS] 가자 [SEP]"

    d = tokenizer.decode([2, 75, 130, 3], skip_special_tokens=True)
    assert d == "가자"

def test_tokenize():
    t = tokenizer.tokenize("맞습니다요!")
    assert t == ["맞/V", "습니다요/E", "!"]

def test_save_pretrained():
    path = tempfile.gettempdir() + '/test_tokenizer'
    tokenizer.save_pretrained(path)
    new_tokenizer = AutoTokenizer.from_pretrained(path)
    assert new_tokenizer.get_vocab() == tokenizer.get_vocab()

if __name__ == '__main__':
    for k, v in locals().copy().items():
        if k.startswith('test'): v()
